# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
import os
import shutil
import sys
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.gsl import find_GSL
from pyomo.common.log import LoggingIntercept
from pyomo.common.tempfiles import TempfileManager
from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Objective,
    Expression,
    SolverFactory,
    value,
    Param,
)
from pyomo.core.base.external import (
    PythonCallbackFunction,
    ExternalFunction,
    AMPLExternalFunction,
)
from pyomo.core.base.units_container import pint_available, units
from pyomo.core.expr.numeric_expr import (
    ExternalFunctionExpression,
    NPV_ExternalFunctionExpression,
)
from pyomo.opt import check_available_solvers


def _count(*args):
    return len(args)


def _sum(*args):
    return 2 + sum(args)


def _f(x, y, z):
    return x**2 + 3 * x * y + x * y * z**2


def _g(args, fixed):
    x, y, z = args[:3]
    return [2 * x + 3 * y + y * z**2, 3 * x + x * z**2, 2 * x * y * z]


def _h(args, fixed):
    x, y, z = args[:3]
    return [2, 3 + z**2, 0, 2 * y * z, 2 * x * z, 2 * x * y]


def _g_bad(args, fixed):
    x, y, z = args[:3]
    return [2 * x + 3 * y + y * z**2, 3 * x + x * z**2, 2 * x * y * z, 0]


def _h_bad(args, fixed):
    x, y, z = args[:3]
    return [
        # 2,
        3 + z**2,
        0,
        2 * y * z,
        2 * x * z,
        2 * x * y,
    ]


def _fgh(args, fixed, fgh):
    return _f(*args), _g(args, fixed), _h(args, fixed)


class TestPythonCallbackFunction(unittest.TestCase):
    def test_constructor_errors(self):
        m = ConcreteModel()
        with self.assertRaisesRegex(
            ValueError,
            "Duplicate definition of external function "
            r"through positional and keyword \('function='\)",
        ):
            m.f = ExternalFunction(_count, function=_count)
        with self.assertRaisesRegex(
            ValueError,
            "PythonCallbackFunction constructor only "
            "supports 0 - 3 positional arguments",
        ):
            m.f = ExternalFunction(1, 2, 3, 4)
        with self.assertRaisesRegex(
            ValueError,
            "Cannot specify 'fgh' with any of {'function', 'gradient', hessian'}",
        ):
            m.f = ExternalFunction(_count, fgh=_fgh)

    def test_call_countArgs(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_count)
        self.assertIsInstance(m.f, PythonCallbackFunction)
        self.assertEqual(value(m.f()), 0)
        self.assertEqual(value(m.f(2)), 1)
        self.assertEqual(value(m.f(2, 3)), 2)

    def test_call_sumfcn(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_sum)
        self.assertIsInstance(m.f, PythonCallbackFunction)
        self.assertEqual(value(m.f()), 2.0)
        self.assertEqual(value(m.f(1)), 3.0)
        self.assertEqual(value(m.f(1, 2)), 5.0)

    def test_evaluate_fgh_fgh(self):
        m = ConcreteModel()
        m.f = ExternalFunction(fgh=_fgh)
        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id))
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(
            g, [2 * 5 + 3 * 7 + 7 * 11**2, 3 * 5 + 5 * 11**2, 2 * 5 * 7 * 11, 0]
        )
        self.assertEqual(
            h, [2, 3 + 11**2, 0, 2 * 7 * 11, 2 * 5 * 11, 2 * 5 * 7, 0, 0, 0, 0]
        )

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fixed=[0, 1, 0, 1])
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(g, [2 * 5 + 3 * 7 + 7 * 11**2, 0, 2 * 5 * 7 * 11, 0])
        self.assertEqual(h, [2, 0, 0, 2 * 7 * 11, 0, 2 * 5 * 7, 0, 0, 0, 0])

    def test_evaluate_fgh_f_g_h(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_f, _g, _h)
        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id))
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(
            g, [2 * 5 + 3 * 7 + 7 * 11**2, 3 * 5 + 5 * 11**2, 2 * 5 * 7 * 11, 0]
        )
        self.assertEqual(
            h, [2, 3 + 11**2, 0, 2 * 7 * 11, 2 * 5 * 11, 2 * 5 * 7, 0, 0, 0, 0]
        )

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fixed=[0, 1, 0, 1])
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(g, [2 * 5 + 3 * 7 + 7 * 11**2, 0, 2 * 5 * 7 * 11, 0])
        self.assertEqual(h, [2, 0, 0, 2 * 7 * 11, 0, 2 * 5 * 7, 0, 0, 0, 0])

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=1)
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(
            g, [2 * 5 + 3 * 7 + 7 * 11**2, 3 * 5 + 5 * 11**2, 2 * 5 * 7 * 11, 0]
        )
        self.assertIsNone(h)

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=0)
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertIsNone(g)
        self.assertIsNone(h)

    def test_evaluate_fgh_f_g(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_f, _g)
        with self.assertRaisesRegex(
            RuntimeError,
            "ExternalFunction 'f' was not defined with a Hessian callback.",
        ):
            f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id))

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=1)
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertEqual(
            g, [2 * 5 + 3 * 7 + 7 * 11**2, 3 * 5 + 5 * 11**2, 2 * 5 * 7 * 11, 0]
        )
        self.assertIsNone(h)

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=0)
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertIsNone(g)
        self.assertIsNone(h)

    def test_evaluate_fgh_f(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_f)
        with self.assertRaisesRegex(
            RuntimeError,
            "ExternalFunction 'f' was not defined with a gradient callback.",
        ):
            f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id))

        with self.assertRaisesRegex(
            RuntimeError,
            "ExternalFunction 'f' was not defined with a gradient callback.",
        ):
            f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=1)

        f, g, h = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=0)
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)
        self.assertIsNone(g)
        self.assertIsNone(h)

    def test_evaluate_errors(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_f, _g_bad, _h_bad)
        f = m.f.evaluate((5, 7, 11, m.f._fcn_id))
        self.assertEqual(f, 5**2 + 3 * 5 * 7 + 5 * 7 * 11**2)

        with self.assertRaisesRegex(
            RuntimeError, "PythonCallbackFunction called with invalid Global ID"
        ):
            f = m.f.evaluate((5, 7, 11, -1))

        with self.assertRaisesRegex(
            RuntimeError,
            "External function 'f' returned an invalid "
            r"derivative vector \(expected 4, received 5\)",
        ):
            f = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=1)

        with self.assertRaisesRegex(
            RuntimeError,
            "External function 'f' returned an invalid "
            r"Hessian matrix \(expected 10, received 9\)",
        ):
            f = m.f.evaluate_fgh((5, 7, 11, m.f._fcn_id), fgh=2)

    def test_getname(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_sum)
        self.assertIsInstance(m.f, PythonCallbackFunction)
        self.assertEqual(m.f.name, "f")
        self.assertEqual(m.f.local_name, "f")
        self.assertEqual(m.f.getname(), "f")
        self.assertEqual(m.f.getname(True), "f")

        M = ConcreteModel()
        M.m = m
        self.assertEqual(M.m.f.name, "m.f")
        self.assertEqual(M.m.f.local_name, "f")
        self.assertEqual(M.m.f.getname(), "f")
        self.assertEqual(M.m.f.getname(True), "m.f")

    def test_extra_kwargs(self):
        m = ConcreteModel()
        with self.assertRaises(ValueError):
            m.f = ExternalFunction(_count, this_should_raise_error='foo')

    def test_clone(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_sum)
        m.x = Var(initialize=3)
        m.y = Var(initialize=5)
        m.e = Expression(expr=m.f(m.x, m.y))
        self.assertIsInstance(m.f, PythonCallbackFunction)
        self.assertEqual(m.f._fcn_id, m.e.arg(0).arg(-1).value)
        self.assertEqual(value(m.e), 10)

        i = m.clone()
        self.assertIsInstance(i.f, PythonCallbackFunction)
        self.assertIsNot(i.f, m.f)
        self.assertIsNot(i.e, m.e)
        self.assertIsNot(i.e.arg(0), m.e.arg(0))
        self.assertEqual(i.f._fcn_id, i.e.arg(0).arg(-1).value)
        self.assertNotEqual(i.f._fcn_id, m.f._fcn_id)
        self.assertEqual(value(i.e), 10)

    def test_partial_clone(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_sum)
        m.x = Var(initialize=3)
        m.y = Var(initialize=5)
        m.b = Block()
        m.b.e = Expression(expr=m.f(m.x, m.y))
        self.assertIsInstance(m.f, PythonCallbackFunction)
        self.assertEqual(m.f._fcn_id, m.b.e.arg(0).arg(-1).value)
        self.assertEqual(value(m.b.e), 10)

        m.c = m.b.clone()
        self.assertIsNot(m.b.e, m.c.e)
        self.assertIsNot(m.b.e.arg(0), m.c.e.arg(0))
        self.assertEqual(m.f._fcn_id, m.b.e.arg(0).arg(-1).value)
        self.assertEqual(m.f._fcn_id, m.c.e.arg(0).arg(-1).value)
        self.assertEqual(value(m.c.e), 10)

        # save / restore should not change the fcn_id
        _fcn_id = m.f._fcn_id
        m.f.__setstate__(m.f.__getstate__())
        self.assertEqual(m.f._fcn_id, _fcn_id)

        self.assertIsNot(m.b.e, m.c.e)
        self.assertIsNot(m.b.e.arg(0), m.c.e.arg(0))
        self.assertEqual(m.f._fcn_id, m.b.e.arg(0).arg(-1).value)
        self.assertEqual(m.f._fcn_id, m.c.e.arg(0).arg(-1).value)
        self.assertEqual(value(m.c.e), 10)

    def test_properties(self):
        m = ConcreteModel()
        m.f = ExternalFunction(_sum)
        e = m.f()
        self.assertIsInstance(e, NPV_ExternalFunctionExpression)
        self.assertFalse(e.is_constant())
        self.assertTrue(e.is_fixed())
        self.assertFalse(e.is_potentially_variable())

        # This is just a check for coverage: the expression is_constant
        # catches that the function ID is a NumericConstant and doesn't
        # call the method.
        self.assertFalse(e.arg(0).is_constant())
        self.assertTrue(e.arg(0).is_fixed())
        self.assertFalse(e.arg(0).is_potentially_variable())

        m.p = Param(initialize=1)
        e = m.f(m.p)
        self.assertIsInstance(e, NPV_ExternalFunctionExpression)
        self.assertFalse(e.is_constant())
        self.assertTrue(e.is_fixed())
        self.assertFalse(e.is_potentially_variable())

        m.x = Var(initialize=1)
        e = m.f(m.p, m.x)
        self.assertIsInstance(e, ExternalFunctionExpression)
        self.assertFalse(e.is_constant())
        self.assertFalse(e.is_fixed())
        self.assertTrue(e.is_potentially_variable())

    def test_pprint(self):
        m = ConcreteModel()
        m.h = ExternalFunction(_count)

        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
1 ExternalFunction Declarations
    h : function=_count, units=None, arg_units=None

1 Declarations: h
        """.strip(),
        )

        if not pint_available:
            return
        m.i = ExternalFunction(
            function=_sum, units=units.kg, arg_units=[units.m, units.s]
        )
        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
2 ExternalFunction Declarations
    h : function=_count, units=None, arg_units=None
    i : function=_sum, units=kg, arg_units=['m', 's']

2 Declarations: h i
        """.strip(),
        )

    def test_pprint(self):
        m = ConcreteModel()
        m.h = ExternalFunction(_g)
        out = StringIO()
        m.pprint()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
1 ExternalFunction Declarations
    h : function=_g, units=None, arg_units=None

1 Declarations: h
        """.strip(),
        )

        if not pint_available:
            return
        m.i = ExternalFunction(
            function=_h, units=units.kg, arg_units=[units.m, units.s]
        )
        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
2 ExternalFunction Declarations
    h : function=_g, units=None, arg_units=None
    i : function=_h, units=kg, arg_units=['m', 's']

2 Declarations: h i
        """.strip(),
        )


class TestAMPLExternalFunction(unittest.TestCase):
    def assertListsAlmostEqual(self, first, second, places=7, msg=None):
        self.assertEqual(len(first), len(second))
        msg = "lists %s and %s differ at item " % (first, second)
        for i, a in enumerate(first):
            self.assertAlmostEqual(a, second[i], places, msg + str(i))

    def test_getname(self):
        m = ConcreteModel()
        m.f = ExternalFunction(library="junk.so", function="junk")
        self.assertIsInstance(m.f, AMPLExternalFunction)
        self.assertEqual(m.f.name, "f")
        self.assertEqual(m.f.local_name, "f")
        self.assertEqual(m.f.getname(), "f")
        self.assertEqual(m.f.getname(True), "f")

        M = ConcreteModel()
        M.m = m
        self.assertEqual(M.m.f.name, "m.f")
        self.assertEqual(M.m.f.local_name, "f")
        self.assertEqual(M.m.f.getname(), "f")
        self.assertEqual(M.m.f.getname(True), "m.f")

    @unittest.skipIf(
        sys.platform.lower().startswith('win'),
        "Cannot (easily) unload a DLL in Windows, so "
        "cannot clean up the 'temporary' DLL",
    )
    def test_load_local_asl_library(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")

        LIB = 'test_pyomo_external_gsl.dll'

        model = ConcreteModel()
        model.gamma = ExternalFunction(library=LIB, function="gsl_sf_gamma")
        model.x = Var(initialize=3, bounds=(1e-5, None))
        model.o = Objective(expr=model.gamma(model.x))

        with TempfileManager.new_context() as tempfile:
            dname = tempfile.mkdtemp()
            shutil.copyfile(DLL, os.path.join(dname, LIB))
            # Without changing directories, the load should fail
            with self.assertRaises(OSError):
                value(model.o)
            # Changing directories should pick up the library
            try:
                orig_dir = os.getcwd()
                os.chdir(dname)
                self.assertAlmostEqual(value(model.o), 2.0, 7)
            finally:
                os.chdir(orig_dir)

    def test_unknown_library(self):
        m = ConcreteModel()
        with LoggingIntercept() as LOG:
            m.ef = ExternalFunction(
                library='unknown_pyomo_external_testing_function', function='f'
            )
        self.assertEqual(
            LOG.getvalue(),
            'Defining AMPL external function, but cannot locate '
            'specified library "unknown_pyomo_external_testing_function"\n',
        )

    def test_eval_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.gamma = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        model.bessel = ExternalFunction(library=DLL, function="gsl_sf_bessel_Jnu")
        model.x = Var(initialize=3, bounds=(1e-5, None))
        model.o = Objective(expr=model.gamma(model.x))
        self.assertAlmostEqual(value(model.o), 2.0, 7)

        f = model.bessel.evaluate((0.5, 2.0))
        self.assertAlmostEqual(f, 0.5130161365618272, 7)

    def test_eval_gsl_error(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.bogus = ExternalFunction(library=DLL, function="bogus_function")
        with self.assertRaisesRegex(
            RuntimeError,
            "Error: external function 'bogus_function' was "
            "not registered within external library(?s:.*)gsl_sf_gamma",
        ):
            f = model.bogus.evaluate((1,))

    def test_eval_fgh_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.gamma = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        model.beta = ExternalFunction(library=DLL, function="gsl_sf_beta")
        model.bessel = ExternalFunction(library=DLL, function="gsl_sf_bessel_Jnu")

        f, g, h = model.gamma.evaluate_fgh((2.0,))
        self.assertAlmostEqual(f, 1.0, 7)
        self.assertListsAlmostEqual(g, [0.422784335098467], 7)
        self.assertListsAlmostEqual(h, [0.8236806608528794], 7)

        f, g, h = model.beta.evaluate_fgh((2.5, 2.0), fixed=[1, 1])
        self.assertAlmostEqual(f, 0.11428571428571432, 7)
        self.assertListsAlmostEqual(g, [0.0, 0.0], 7)
        self.assertListsAlmostEqual(h, [0.0, 0.0, 0.0], 7)

        f, g, h = model.beta.evaluate_fgh((2.5, 2.0), fixed=[0, 1])
        self.assertAlmostEqual(f, 0.11428571428571432, 7)
        self.assertListsAlmostEqual(g, [-0.07836734693877555, 0.0], 7)
        self.assertListsAlmostEqual(h, [0.08135276967930034, 0.0, 0.0], 7)

        f, g, h = model.beta.evaluate_fgh((2.5, 2.0))
        self.assertAlmostEqual(f, 0.11428571428571432, 7)
        self.assertListsAlmostEqual(g, [-0.07836734693877555, -0.11040989614412142], 7)
        self.assertListsAlmostEqual(
            h, [0.08135276967930034, 0.0472839170086535, 0.15194654464270113], 7
        )

        f, g, h = model.beta.evaluate_fgh((2.5, 2.0), fgh=1)
        self.assertAlmostEqual(f, 0.11428571428571432, 7)
        self.assertListsAlmostEqual(g, [-0.07836734693877555, -0.11040989614412142], 7)
        self.assertIsNone(h)

        f, g, h = model.beta.evaluate_fgh((2.5, 2.0), fgh=0)
        self.assertAlmostEqual(f, 0.11428571428571432, 7)
        self.assertIsNone(g)
        self.assertIsNone(h)

        f, g, h = model.bessel.evaluate_fgh((2.5, 2.0), fixed=[1, 0])
        self.assertAlmostEqual(f, 0.223924531469, 7)
        self.assertListsAlmostEqual(g, [0.0, 0.21138811435101745], 7)
        self.assertListsAlmostEqual(h, [0.0, 0.0, 0.02026349177575621], 7)

        # Note: Not all AMPL-GSL functions honor the fixed flag
        # (notably, gamma and bessel do not as of 12/2021).  We will
        # test that our interface corrects that

        f, g, h = model.gamma.evaluate_fgh((2.0,), fixed=[1])
        self.assertAlmostEqual(f, 1.0, 7)
        self.assertListsAlmostEqual(g, [0.0], 7)
        self.assertListsAlmostEqual(h, [0.0], 7)

        f, g, h = model.bessel.evaluate_fgh((2.5, 2.0), fixed=[1, 1])
        self.assertAlmostEqual(f, 0.223924531469, 7)
        self.assertListsAlmostEqual(g, [0.0, 0.0], 7)
        self.assertListsAlmostEqual(h, [0.0, 0.0, 0.0], 7)

    @unittest.skipIf(
        not check_available_solvers('ipopt'), "The 'ipopt' solver is not available"
    )
    def test_solve_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.z_func = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        model.x = Var(initialize=3, bounds=(1e-5, None))
        model.o = Objective(expr=model.z_func(model.x))
        opt = SolverFactory('ipopt')
        res = opt.solve(model, tee=True)
        self.assertAlmostEqual(value(model.o), 0.885603194411, 7)

    @unittest.skipIf(
        not check_available_solvers('ipopt'), "The 'ipopt' solver is not available"
    )
    def test_solve_gsl_function_const_arg(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.z_func = ExternalFunction(library=DLL, function="gsl_sf_beta")
        model.x = Var(initialize=1, bounds=(0.1, None))
        model.o = Objective(expr=-model.z_func(1, model.x))
        opt = SolverFactory('ipopt')
        res = opt.solve(model, tee=True)
        self.assertAlmostEqual(value(model.x), 0.1, 5)

    @unittest.skipIf(
        not check_available_solvers('ipopt_v2'),
        "The 'ipopt_v2' solver is not available",
    )
    def test_solve_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.z_func = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        model.x = Var(initialize=3, bounds=(1e-5, None))
        model.o = Objective(expr=model.z_func(model.x))
        opt = SolverFactory('ipopt_v2')
        res = opt.solve(model, tee=True)
        self.assertAlmostEqual(value(model.o), 0.885603194411, 7)

    @unittest.skipIf(
        not check_available_solvers('ipopt_v2'),
        "The 'ipopt_v2' solver is not available",
    )
    def test_solve_gsl_function_const_arg(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.z_func = ExternalFunction(library=DLL, function="gsl_sf_beta")
        model.x = Var(initialize=1, bounds=(0.1, None))
        model.o = Objective(expr=-model.z_func(1, model.x))
        opt = SolverFactory('ipopt_v2')
        res = opt.solve(model, tee=True)
        self.assertAlmostEqual(value(model.x), 0.1, 5)

    @unittest.skipIf(
        not check_available_solvers('ipopt'), "The 'ipopt' solver is not available"
    )
    def test_clone_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        m = ConcreteModel()
        m.z_func = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        self.assertIsInstance(m.z_func, AMPLExternalFunction)
        m.x = Var(initialize=3, bounds=(1e-5, None))
        m.o = Objective(expr=m.z_func(m.x))

        opt = SolverFactory('ipopt')

        # Test a simple clone...
        model2 = m.clone()
        res = opt.solve(model2, tee=True)
        self.assertAlmostEqual(value(model2.o), 0.885603194411, 7)

        # Trigger the library to be loaded.  This tests that the CDLL
        # objects that are created when the SO/DLL are loaded do not
        # interfere with cloning the model.
        self.assertAlmostEqual(value(m.o), 2)
        model3 = m.clone()
        res = opt.solve(model3, tee=True)
        self.assertAlmostEqual(value(model3.o), 0.885603194411, 7)

    def test_pprint(self):
        m = ConcreteModel()
        m.f = ExternalFunction(library="junk.so", function="junk")

        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
1 ExternalFunction Declarations
    f : function=junk, library=junk.so, units=None, arg_units=None

1 Declarations: f
        """.strip(),
        )

        if not pint_available:
            return
        m.g = ExternalFunction(
            library="junk.so",
            function="junk",
            units=units.kg,
            arg_units=[units.m, units.s],
        )
        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
2 ExternalFunction Declarations
    f : function=junk, library=junk.so, units=None, arg_units=None
    g : function=junk, library=junk.so, units=kg, arg_units=['m', 's']

2 Declarations: f g
        """.strip(),
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.contrib.piecewise.tests.models as models
from pyomo.core import Var
from pyomo.core.base import TransformationFactory
from pyomo.environ import value
from pyomo.gdp import Disjunct, Disjunction


def check_trans_block_structure(test, block):
    # One (indexed) disjunct
    test.assertEqual(len(block.component_map(Disjunct)), 1)
    # One disjunction
    test.assertEqual(len(block.component_map(Disjunction)), 1)
    # The 'z' var (that we will substitute in for the function being
    # approximated) is here:
    test.assertEqual(len(block.component_map(Var)), 1)
    test.assertIsInstance(block.substitute_var, Var)


def check_log_x_model_soln(test, m):
    test.assertAlmostEqual(value(m.x), 4)
    test.assertAlmostEqual(value(m.x1), 1)
    test.assertAlmostEqual(value(m.x2), 1)
    test.assertAlmostEqual(value(m.obj), m.f2(4))


def check_transformation_do_not_descend(test, transformation, m=None):
    if m is None:
        m = models.make_log_x_model()
    transform = TransformationFactory(transformation)
    transform.apply_to(m)

    test.check_pw_log(m)
    test.check_pw_paraboloid(m)


def check_transformation_PiecewiseLinearFunction_targets(test, transformation, m=None):
    if m is None:
        m = models.make_log_x_model()
    transform = TransformationFactory(transformation)
    transform.apply_to(m, targets=[m.pw_log])

    test.check_pw_log(m)

    # And check that the paraboloid was *not* transformed.
    test.assertIsNone(m.pw_paraboloid.get_transformation_var(m.paraboloid_expr))


def check_descend_into_expressions(test, transformation, m=None):
    if m is None:
        m = models.make_log_x_model()
    transform = TransformationFactory(transformation)
    transform.apply_to(m, descend_into_expressions=True)

    # Everything should be transformed
    test.check_pw_log(m)
    test.check_pw_paraboloid(m)


def check_descend_into_expressions_constraint_target(test, transformation, m=None):
    if m is None:
        m = models.make_log_x_model()
    transform = TransformationFactory(transformation)
    transform.apply_to(m, descend_into_expressions=True, targets=[m.indexed_c])

    test.check_pw_paraboloid(m)
    # And check that the log was *not* transformed.
    test.assertIsNone(m.pw_log.get_transformation_var(m.log_expr))


def check_descend_into_expressions_objective_target(test, transformation, m=None):
    if m is None:
        m = models.make_log_x_model()
    transform = TransformationFactory(transformation)
    transform.apply_to(m, descend_into_expressions=True, targets=[m.obj])

    test.check_pw_log(m)
    # And check that the paraboloid was *not* transformed.
    test.assertIsNone(m.pw_paraboloid.get_transformation_var(m.paraboloid_expr))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly modifies product expressions
#          with only constant terms in the denominator (that
#          are involved in nonlinear expressions).
#          The ASL differentiation routines seem to have a
#          bug that causes the lagrangian hessian to become
#          dense unless this constant term in moved to the
#          numerator.
#
#          This test model relies on the gjh_asl_json executable. It
#          will not solve if sent to a real optimizer.
#

from pyomo.environ import ConcreteModel, Var, Param, Objective, Constraint

model = ConcreteModel()
model.x = Var(bounds=(-1.0, 1.0), initialize=1.0)
model.y = Var(bounds=(-1.0, 1.0), initialize=2.0)
model.v = Var(bounds=(-1.0, 1.0), initialize=3.0)
model.p = Var(initialize=2.0)
model.p.fixed = True
model.q = Param(initialize=2.0)

model.OBJ = Objective(expr=model.x)
model.CON1a = Constraint(
    expr=1.0 / model.p / model.q * model.v * (model.x - model.y) == 2.0
)
model.CON2a = Constraint(
    expr=model.v * 1.0 / model.p / model.p * (model.x - model.y) == 2.0
)
model.CON3a = Constraint(expr=model.v * (model.x - model.y) / model.p / model.q == 2.0)
model.CON4a = Constraint(
    expr=model.v * (model.x / model.p / model.q - model.y / model.p / model.q) == 2.0
)
model.CON5a = Constraint(
    expr=model.v * (model.x - model.y) * (1.0 / model.p / model.q) == 2.0
)
model.CON6a = Constraint(
    expr=model.v * (model.x - model.y) - 2.0 * model.p * model.q == 0
)

model.CON1b = Constraint(
    expr=1.0 / (model.p * model.q) * model.v * (model.x - model.y) == 2.0
)
model.CON2b = Constraint(
    expr=model.v * 1.0 / (model.p * model.p) * (model.x - model.y) == 2.0
)
model.CON3b = Constraint(
    expr=model.v * (model.x - model.y) / (model.p * model.q) == 2.0
)
model.CON4b = Constraint(
    expr=model.v * (model.x / (model.p * model.q) - model.y / (model.p * model.q))
    == 2.0
)
model.CON5b = Constraint(
    expr=model.v * (model.x - model.y) * (1.0 / (model.p * model.q)) == 2.0
)
model.CON6b = Constraint(
    expr=model.v * (model.x - model.y) - 2.0 * (model.p * model.q) == 0
)

model.CON1c = Constraint(
    expr=1.0 / (model.p + model.q) * model.v * (model.x - model.y) == 2.0
)
model.CON2c = Constraint(
    expr=model.v * 1.0 / (model.p + model.p) * (model.x - model.y) == 2.0
)
model.CON3c = Constraint(
    expr=model.v * (model.x - model.y) / (model.p + model.q) == 2.0
)
model.CON4c = Constraint(
    expr=model.v * (model.x / (model.p + model.q) - model.y / (model.p + model.q))
    == 2.0
)
model.CON5c = Constraint(
    expr=model.v * (model.x - model.y) * (1.0 / (model.p + model.q)) == 2.0
)
model.CON6c = Constraint(
    expr=model.v * (model.x - model.y) == 2.0 * (model.p + model.q)
)

model.CON1d = Constraint(
    expr=1.0 / ((model.p + model.q) ** 2) * model.v * (model.x - model.y) == 2.0
)
model.CON2d = Constraint(
    expr=model.v * 1.0 / ((model.p + model.p) ** 2) * (model.x - model.y) == 2.0
)
model.CON3d = Constraint(
    expr=model.v * (model.x - model.y) / ((model.p + model.q) ** 2) == 2.0
)
model.CON4d = Constraint(
    expr=model.v
    * (model.x / ((model.p + model.q) ** 2) - model.y / ((model.p + model.q) ** 2))
    == 2.0
)
model.CON5d = Constraint(
    expr=model.v * (model.x - model.y) * (1.0 / ((model.p + model.q) ** 2)) == 2.0
)
model.CON6d = Constraint(
    expr=model.v * (model.x - model.y) - 2.0 * ((model.p + model.q) ** 2) == 0
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd, pandas_available

uuid_available = True
try:
    import uuid
except:
    uuid_available = False

import pyomo.common.unittest as unittest
import os
import pyomo.contrib.parmest.parmest as parmest
import pyomo.contrib.parmest.scenariocreator as sc
import pyomo.environ as pyo
from pyomo.environ import SolverFactory

ipopt_available = SolverFactory("ipopt").available()

testdir = os.path.dirname(os.path.abspath(__file__))


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestScenarioReactorDesign(unittest.TestCase):
    def setUp(self):
        from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
            ReactorDesignExperiment,
        )

        # Data from the design
        data = pd.DataFrame(
            data=[
                [1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
                [1.20, 10000, 3680.7, 1070.0, 1486.1, 1881.6],
                [1.25, 10000, 3750.0, 1071.4, 1428.6, 1875.0],
                [1.30, 10000, 3817.1, 1072.2, 1374.6, 1868.0],
                [1.35, 10000, 3882.2, 1072.4, 1324.0, 1860.7],
                [1.40, 10000, 3945.4, 1072.1, 1276.3, 1853.1],
                [1.45, 10000, 4006.7, 1071.3, 1231.4, 1845.3],
                [1.50, 10000, 4066.4, 1070.1, 1189.0, 1837.3],
                [1.55, 10000, 4124.4, 1068.5, 1148.9, 1829.1],
                [1.60, 10000, 4180.9, 1066.5, 1111.0, 1820.8],
                [1.65, 10000, 4235.9, 1064.3, 1075.0, 1812.4],
                [1.70, 10000, 4289.5, 1061.8, 1040.9, 1803.9],
                [1.75, 10000, 4341.8, 1059.0, 1008.5, 1795.3],
                [1.80, 10000, 4392.8, 1056.0, 977.7, 1786.7],
                [1.85, 10000, 4442.6, 1052.8, 948.4, 1778.1],
                [1.90, 10000, 4491.3, 1049.4, 920.5, 1769.4],
                [1.95, 10000, 4538.8, 1045.8, 893.9, 1760.8],
            ],
            columns=["sv", "caf", "ca", "cb", "cc", "cd"],
        )

        # Create an experiment list
        exp_list = []
        for i in range(data.shape[0]):
            exp_list.append(ReactorDesignExperiment(data, i))

        self.pest = parmest.Estimator(exp_list, obj_function='SSE')

    def test_scen_from_exps(self):
        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        experimentscens = sc.ScenarioSet("Experiments")
        scenmaker.ScenariosFromExperiments(experimentscens)
        experimentscens.write_csv("delme_exp_csv.csv")
        df = pd.read_csv("delme_exp_csv.csv")
        os.remove("delme_exp_csv.csv")
        # March '20: all reactor_design experiments have the same theta values!
        k1val = df.loc[5].at["k1"]
        self.assertAlmostEqual(k1val, 5.0 / 6.0, places=2)
        tval = experimentscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 5.0 / 6.0, places=2)

    @unittest.skipIf(not uuid_available, "The uuid module is not available")
    def test_no_csv_if_empty(self):
        # low level test of scenario sets
        # verify that nothing is written, but no errors with empty set

        emptyset = sc.ScenarioSet("empty")
        tfile = uuid.uuid4().hex + ".csv"
        emptyset.write_csv(tfile)
        self.assertFalse(
            os.path.exists(tfile), "ScenarioSet wrote csv in spite of empty set"
        )


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestScenarioSemibatch(unittest.TestCase):
    def setUp(self):
        import pyomo.contrib.parmest.examples.semibatch.semibatch as sb
        import json

        self.fbase = os.path.join(testdir, "..", "examples", "semibatch")
        # Data, list of dictionaries
        data = []
        for exp_num in range(10):
            fname = "exp" + str(exp_num + 1) + ".out"
            fullname = os.path.join(self.fbase, fname)
            with open(fullname, "r") as infile:
                d = json.load(infile)
                data.append(d)

        # Note, the model already includes a 'SecondStageCost' expression
        # for the sum of squared error that will be used in parameter estimation

        # Create an experiment list
        exp_list = []
        for i in range(len(data)):
            exp_list.append(sb.SemiBatchExperiment(data[i]))

        self.pest = parmest.Estimator(exp_list)

    def test_semibatch_bootstrap(self):
        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        bootscens = sc.ScenarioSet("Bootstrap")
        numtomake = 2
        scenmaker.ScenariosFromBootstrap(bootscens, numtomake, seed=1134)
        tval = bootscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 20.64, places=1)


###########################
# tests for deprecated UI #
###########################


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestScenarioReactorDesignDeprecated(unittest.TestCase):
    def setUp(self):

        def reactor_design_model(data):
            # Create the concrete model
            model = pyo.ConcreteModel()

            # Rate constants
            model.k1 = pyo.Param(
                initialize=5.0 / 6.0, within=pyo.PositiveReals, mutable=True
            )  # min^-1
            model.k2 = pyo.Param(
                initialize=5.0 / 3.0, within=pyo.PositiveReals, mutable=True
            )  # min^-1
            model.k3 = pyo.Param(
                initialize=1.0 / 6000.0, within=pyo.PositiveReals, mutable=True
            )  # m^3/(gmol min)

            # Inlet concentration of A, gmol/m^3
            if isinstance(data, dict) or isinstance(data, pd.Series):
                model.caf = pyo.Param(
                    initialize=float(data["caf"]), within=pyo.PositiveReals
                )
            elif isinstance(data, pd.DataFrame):
                model.caf = pyo.Param(
                    initialize=float(data.iloc[0]["caf"]), within=pyo.PositiveReals
                )
            else:
                raise ValueError("Unrecognized data type.")

            # Space velocity (flowrate/volume)
            if isinstance(data, dict) or isinstance(data, pd.Series):
                model.sv = pyo.Param(
                    initialize=float(data["sv"]), within=pyo.PositiveReals
                )
            elif isinstance(data, pd.DataFrame):
                model.sv = pyo.Param(
                    initialize=float(data.iloc[0]["sv"]), within=pyo.PositiveReals
                )
            else:
                raise ValueError("Unrecognized data type.")

            # Outlet concentration of each component
            model.ca = pyo.Var(initialize=5000.0, within=pyo.PositiveReals)
            model.cb = pyo.Var(initialize=2000.0, within=pyo.PositiveReals)
            model.cc = pyo.Var(initialize=2000.0, within=pyo.PositiveReals)
            model.cd = pyo.Var(initialize=1000.0, within=pyo.PositiveReals)

            # Objective
            model.obj = pyo.Objective(expr=model.cb, sense=pyo.maximize)

            # Constraints
            model.ca_bal = pyo.Constraint(
                expr=(
                    0
                    == model.sv * model.caf
                    - model.sv * model.ca
                    - model.k1 * model.ca
                    - 2.0 * model.k3 * model.ca**2.0
                )
            )

            model.cb_bal = pyo.Constraint(
                expr=(
                    0
                    == -model.sv * model.cb + model.k1 * model.ca - model.k2 * model.cb
                )
            )

            model.cc_bal = pyo.Constraint(
                expr=(0 == -model.sv * model.cc + model.k2 * model.cb)
            )

            model.cd_bal = pyo.Constraint(
                expr=(0 == -model.sv * model.cd + model.k3 * model.ca**2.0)
            )

            return model

        # Data from the design
        data = pd.DataFrame(
            data=[
                [1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
                [1.20, 10000, 3680.7, 1070.0, 1486.1, 1881.6],
                [1.25, 10000, 3750.0, 1071.4, 1428.6, 1875.0],
                [1.30, 10000, 3817.1, 1072.2, 1374.6, 1868.0],
                [1.35, 10000, 3882.2, 1072.4, 1324.0, 1860.7],
                [1.40, 10000, 3945.4, 1072.1, 1276.3, 1853.1],
                [1.45, 10000, 4006.7, 1071.3, 1231.4, 1845.3],
                [1.50, 10000, 4066.4, 1070.1, 1189.0, 1837.3],
                [1.55, 10000, 4124.4, 1068.5, 1148.9, 1829.1],
                [1.60, 10000, 4180.9, 1066.5, 1111.0, 1820.8],
                [1.65, 10000, 4235.9, 1064.3, 1075.0, 1812.4],
                [1.70, 10000, 4289.5, 1061.8, 1040.9, 1803.9],
                [1.75, 10000, 4341.8, 1059.0, 1008.5, 1795.3],
                [1.80, 10000, 4392.8, 1056.0, 977.7, 1786.7],
                [1.85, 10000, 4442.6, 1052.8, 948.4, 1778.1],
                [1.90, 10000, 4491.3, 1049.4, 920.5, 1769.4],
                [1.95, 10000, 4538.8, 1045.8, 893.9, 1760.8],
            ],
            columns=["sv", "caf", "ca", "cb", "cc", "cd"],
        )

        theta_names = ["k1", "k2", "k3"]

        def SSE(model, data):
            expr = (
                (float(data.iloc[0]["ca"]) - model.ca) ** 2
                + (float(data.iloc[0]["cb"]) - model.cb) ** 2
                + (float(data.iloc[0]["cc"]) - model.cc) ** 2
                + (float(data.iloc[0]["cd"]) - model.cd) ** 2
            )
            return expr

        self.pest = parmest.Estimator(reactor_design_model, data, theta_names, SSE)

    def test_scen_from_exps(self):
        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        experimentscens = sc.ScenarioSet("Experiments")
        scenmaker.ScenariosFromExperiments(experimentscens)
        experimentscens.write_csv("delme_exp_csv.csv")
        df = pd.read_csv("delme_exp_csv.csv")
        os.remove("delme_exp_csv.csv")
        # March '20: all reactor_design experiments have the same theta values!
        k1val = df.loc[5].at["k1"]
        self.assertAlmostEqual(k1val, 5.0 / 6.0, places=2)
        tval = experimentscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 5.0 / 6.0, places=2)

    @unittest.skipIf(not uuid_available, "The uuid module is not available")
    def test_no_csv_if_empty(self):
        # low level test of scenario sets
        # verify that nothing is written, but no errors with empty set

        emptyset = sc.ScenarioSet("empty")
        tfile = uuid.uuid4().hex + ".csv"
        emptyset.write_csv(tfile)
        self.assertFalse(
            os.path.exists(tfile), "ScenarioSet wrote csv in spite of empty set"
        )


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestScenarioSemibatchDeprecated(unittest.TestCase):
    def setUp(self):

        import json
        from pyomo.environ import (
            ConcreteModel,
            Set,
            Param,
            Var,
            Constraint,
            ConstraintList,
            Expression,
            Objective,
            TransformationFactory,
            SolverFactory,
            exp,
            minimize,
        )
        from pyomo.dae import ContinuousSet, DerivativeVar

        def generate_model(data):
            # if data is a file name, then load file first
            if isinstance(data, str):
                file_name = data
                try:
                    with open(file_name, "r") as infile:
                        data = json.load(infile)
                except:
                    raise RuntimeError(f"Could not read {file_name} as json")

            # unpack and fix the data
            cameastemp = data["Ca_meas"]
            cbmeastemp = data["Cb_meas"]
            ccmeastemp = data["Cc_meas"]
            trmeastemp = data["Tr_meas"]

            cameas = {}
            cbmeas = {}
            ccmeas = {}
            trmeas = {}
            for i in cameastemp.keys():
                cameas[float(i)] = cameastemp[i]
                cbmeas[float(i)] = cbmeastemp[i]
                ccmeas[float(i)] = ccmeastemp[i]
                trmeas[float(i)] = trmeastemp[i]

            m = ConcreteModel()

            #
            # Measurement Data
            #
            m.measT = Set(initialize=sorted(cameas.keys()))
            m.Ca_meas = Param(m.measT, initialize=cameas)
            m.Cb_meas = Param(m.measT, initialize=cbmeas)
            m.Cc_meas = Param(m.measT, initialize=ccmeas)
            m.Tr_meas = Param(m.measT, initialize=trmeas)

            #
            # Parameters for semi-batch reactor model
            #
            m.R = Param(initialize=8.314)  # kJ/kmol/K
            m.Mwa = Param(initialize=50.0)  # kg/kmol
            m.rhor = Param(initialize=1000.0)  # kg/m^3
            m.cpr = Param(initialize=3.9)  # kJ/kg/K
            m.Tf = Param(initialize=300)  # K
            m.deltaH1 = Param(initialize=-40000.0)  # kJ/kmol
            m.deltaH2 = Param(initialize=-50000.0)  # kJ/kmol
            m.alphaj = Param(initialize=0.8)  # kJ/s/m^2/K
            m.alphac = Param(initialize=0.7)  # kJ/s/m^2/K
            m.Aj = Param(initialize=5.0)  # m^2
            m.Ac = Param(initialize=3.0)  # m^2
            m.Vj = Param(initialize=0.9)  # m^3
            m.Vc = Param(initialize=0.07)  # m^3
            m.rhow = Param(initialize=700.0)  # kg/m^3
            m.cpw = Param(initialize=3.1)  # kJ/kg/K
            m.Ca0 = Param(initialize=data["Ca0"])  # kmol/m^3)
            m.Cb0 = Param(initialize=data["Cb0"])  # kmol/m^3)
            m.Cc0 = Param(initialize=data["Cc0"])  # kmol/m^3)
            m.Tr0 = Param(initialize=300.0)  # K
            m.Vr0 = Param(initialize=1.0)  # m^3

            m.time = ContinuousSet(
                bounds=(0, 21600), initialize=m.measT
            )  # Time in seconds

            #
            # Control Inputs
            #
            def _initTc(m, t):
                if t < 10800:
                    return data["Tc1"]
                else:
                    return data["Tc2"]

            m.Tc = Param(
                m.time, initialize=_initTc, default=_initTc
            )  # bounds= (288,432) Cooling coil temp, control input

            def _initFa(m, t):
                if t < 10800:
                    return data["Fa1"]
                else:
                    return data["Fa2"]

            m.Fa = Param(
                m.time, initialize=_initFa, default=_initFa
            )  # bounds=(0,0.05) Inlet flow rate, control input

            #
            # Parameters being estimated
            #
            m.k1 = Var(initialize=14, bounds=(2, 100))  # 1/s Actual: 15.01
            m.k2 = Var(initialize=90, bounds=(2, 150))  # 1/s Actual: 85.01
            m.E1 = Var(
                initialize=27000.0, bounds=(25000, 40000)
            )  # kJ/kmol Actual: 30000
            m.E2 = Var(
                initialize=45000.0, bounds=(35000, 50000)
            )  # kJ/kmol Actual: 40000
            # m.E1.fix(30000)
            # m.E2.fix(40000)

            #
            # Time dependent variables
            #
            m.Ca = Var(m.time, initialize=m.Ca0, bounds=(0, 25))
            m.Cb = Var(m.time, initialize=m.Cb0, bounds=(0, 25))
            m.Cc = Var(m.time, initialize=m.Cc0, bounds=(0, 25))
            m.Vr = Var(m.time, initialize=m.Vr0)
            m.Tr = Var(m.time, initialize=m.Tr0)
            m.Tj = Var(
                m.time, initialize=310.0, bounds=(288, None)
            )  # Cooling jacket temp, follows coil temp until failure

            #
            # Derivatives in the model
            #
            m.dCa = DerivativeVar(m.Ca)
            m.dCb = DerivativeVar(m.Cb)
            m.dCc = DerivativeVar(m.Cc)
            m.dVr = DerivativeVar(m.Vr)
            m.dTr = DerivativeVar(m.Tr)

            #
            # Differential Equations in the model
            #

            def _dCacon(m, t):
                if t == 0:
                    return Constraint.Skip
                return (
                    m.dCa[t]
                    == m.Fa[t] / m.Vr[t] - m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[t]
                )

            m.dCacon = Constraint(m.time, rule=_dCacon)

            def _dCbcon(m, t):
                if t == 0:
                    return Constraint.Skip
                return (
                    m.dCb[t]
                    == m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[t]
                    - m.k2 * exp(-m.E2 / (m.R * m.Tr[t])) * m.Cb[t]
                )

            m.dCbcon = Constraint(m.time, rule=_dCbcon)

            def _dCccon(m, t):
                if t == 0:
                    return Constraint.Skip
                return m.dCc[t] == m.k2 * exp(-m.E2 / (m.R * m.Tr[t])) * m.Cb[t]

            m.dCccon = Constraint(m.time, rule=_dCccon)

            def _dVrcon(m, t):
                if t == 0:
                    return Constraint.Skip
                return m.dVr[t] == m.Fa[t] * m.Mwa / m.rhor

            m.dVrcon = Constraint(m.time, rule=_dVrcon)

            def _dTrcon(m, t):
                if t == 0:
                    return Constraint.Skip
                return m.rhor * m.cpr * m.dTr[t] == m.Fa[t] * m.Mwa * m.cpr / m.Vr[
                    t
                ] * (m.Tf - m.Tr[t]) - m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[
                    t
                ] * m.deltaH1 - m.k2 * exp(
                    -m.E2 / (m.R * m.Tr[t])
                ) * m.Cb[
                    t
                ] * m.deltaH2 + m.alphaj * m.Aj / m.Vr0 * (
                    m.Tj[t] - m.Tr[t]
                ) + m.alphac * m.Ac / m.Vr0 * (
                    m.Tc[t] - m.Tr[t]
                )

            m.dTrcon = Constraint(m.time, rule=_dTrcon)

            def _singlecooling(m, t):
                return m.Tc[t] == m.Tj[t]

            m.singlecooling = Constraint(m.time, rule=_singlecooling)

            # Initial Conditions
            def _initcon(m):
                yield m.Ca[m.time.first()] == m.Ca0
                yield m.Cb[m.time.first()] == m.Cb0
                yield m.Cc[m.time.first()] == m.Cc0
                yield m.Vr[m.time.first()] == m.Vr0
                yield m.Tr[m.time.first()] == m.Tr0

            m.initcon = ConstraintList(rule=_initcon)

            #
            # Stage-specific cost computations
            #
            def ComputeFirstStageCost_rule(model):
                return 0

            m.FirstStageCost = Expression(rule=ComputeFirstStageCost_rule)

            def AllMeasurements(m):
                return sum(
                    (m.Ca[t] - m.Ca_meas[t]) ** 2
                    + (m.Cb[t] - m.Cb_meas[t]) ** 2
                    + (m.Cc[t] - m.Cc_meas[t]) ** 2
                    + 0.01 * (m.Tr[t] - m.Tr_meas[t]) ** 2
                    for t in m.measT
                )

            def MissingMeasurements(m):
                if data["experiment"] == 1:
                    return sum(
                        (m.Ca[t] - m.Ca_meas[t]) ** 2
                        + (m.Cb[t] - m.Cb_meas[t]) ** 2
                        + (m.Cc[t] - m.Cc_meas[t]) ** 2
                        + (m.Tr[t] - m.Tr_meas[t]) ** 2
                        for t in m.measT
                    )
                elif data["experiment"] == 2:
                    return sum((m.Tr[t] - m.Tr_meas[t]) ** 2 for t in m.measT)
                else:
                    return sum(
                        (m.Cb[t] - m.Cb_meas[t]) ** 2 + (m.Tr[t] - m.Tr_meas[t]) ** 2
                        for t in m.measT
                    )

            m.SecondStageCost = Expression(rule=MissingMeasurements)

            def total_cost_rule(model):
                return model.FirstStageCost + model.SecondStageCost

            m.Total_Cost_Objective = Objective(rule=total_cost_rule, sense=minimize)

            # Discretize model
            disc = TransformationFactory("dae.collocation")
            disc.apply_to(m, nfe=20, ncp=4)
            return m

        # Vars to estimate in parmest
        theta_names = ["k1", "k2", "E1", "E2"]

        self.fbase = os.path.join(testdir, "..", "examples", "semibatch")
        # Data, list of dictionaries
        data = []
        for exp_num in range(10):
            fname = "exp" + str(exp_num + 1) + ".out"
            fullname = os.path.join(self.fbase, fname)
            with open(fullname, "r") as infile:
                d = json.load(infile)
                data.append(d)

        # Note, the model already includes a 'SecondStageCost' expression
        # for the sum of squared error that will be used in parameter estimation

        self.pest = parmest.Estimator(generate_model, data, theta_names)

    def test_semibatch_bootstrap(self):
        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        bootscens = sc.ScenarioSet("Bootstrap")
        numtomake = 2
        scenmaker.ScenariosFromBootstrap(bootscens, numtomake, seed=1134)
        tval = bootscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 20.64, places=1)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.gdpopt.util import time_code, get_main_elapsed_time
from pyomo.contrib.mindtpy.util import calc_jacobians
from pyomo.core import ConstraintList
from pyomo.opt import SolverFactory
from pyomo.contrib.mindtpy.config_options import _get_MindtPy_ECP_config
from pyomo.contrib.mindtpy.algorithm_base_class import _MindtPyAlgorithm
from pyomo.contrib.mindtpy.cut_generation import add_ecp_cuts
from pyomo.opt import TerminationCondition as tc


@SolverFactory.register(
    'mindtpy.ecp', doc='MindtPy: Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo'
)
class MindtPy_ECP_Solver(_MindtPyAlgorithm):
    """
    Decomposition solver for Mixed-Integer Nonlinear Programming (MINLP) problems.

    The MindtPy (Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo) solver
    applies a variety of decomposition-based approaches to solve Mixed-Integer
    Nonlinear Programming (MINLP) problems.
    This class includes:

    - Extended Cutting Plane (ECP)
    """

    CONFIG = _get_MindtPy_ECP_config()

    def MindtPy_iteration_loop(self):
        """Main loop for MindtPy Algorithms.

        This is the outermost function for the Extended Cutting Plane algorithm in this package; this function controls the progress of
        solving the model.

        Raises
        ------
        ValueError
            The strategy value is not correct or not included.
        """
        while self.mip_iter < self.config.iteration_limit:
            # solve MIP main problem
            main_mip, main_mip_results = self.solve_main()

            if self.handle_main_mip_termination(main_mip, main_mip_results):
                break

            # Call the MIP post-solve callback
            with time_code(self.timing, 'Call after main solve'):
                self.config.call_after_main_solve(main_mip)

            if self.algorithm_should_terminate():
                self.last_iter_cuts = False
                break

            add_ecp_cuts(self.mip, self.jacobians, self.config, self.timing)

        self.config.logger.info(
            ' ==============================================================================================='
        )

    def check_config(self):
        config = self.config
        # if ecp tolerance is not provided use bound tolerance
        if config.ecp_tolerance is None:
            config.ecp_tolerance = config.absolute_bound_tolerance
        super().check_config()

    def initialize_mip_problem(self):
        """Deactivate the nonlinear constraints to create the MIP problem."""
        super().initialize_mip_problem()
        self.jacobians = calc_jacobians(
            self.mip.MindtPy_utils.nonlinear_constraint_list,
            self.config.differentiate_mode,
        )  # preload jacobians
        self.mip.MindtPy_utils.cuts.ecp_cuts = ConstraintList(
            doc='Extended Cutting Planes'
        )

    def init_rNLP(self):
        """Initialize the problem by solving the relaxed NLP and then store the optimal variable
        values obtained from solving the rNLP.

        Raises
        ------
        ValueError
            MindtPy unable to handle the termination condition of the relaxed NLP.
        """
        super().init_rNLP(add_oa_cuts=False)

    def algorithm_should_terminate(self):
        """Checks if the algorithm should terminate at the given point.

        This function determines whether the algorithm should terminate based on the solver options and progress.
        (Sets the self.results.solver.termination_condition to the appropriate condition, i.e. optimal,
        maxIterations, maxTimeLimit).

        Returns
        -------
        bool
            True if the algorithm should terminate, False otherwise.
        """
        if self.should_terminate:
            if self.primal_bound == self.primal_bound_progress[0]:
                self.results.solver.termination_condition = tc.noSolution
            else:
                self.results.solver.termination_condition = tc.feasible
            return True

        return (
            self.bounds_converged()
            or self.reached_iteration_limit()
            or self.reached_time_limit()
            or self.reached_stalling_limit()
            or self.all_nonlinear_constraint_satisfied()
        )

    def all_nonlinear_constraint_satisfied(self):
        # check to see if the nonlinear constraints are satisfied
        config = self.config
        MindtPy = self.mip.MindtPy_utils
        nonlinear_constraints = [c for c in MindtPy.nonlinear_constraint_list]
        for nlc in nonlinear_constraints:
            if nlc.has_lb():
                try:
                    lower_slack = nlc.lslack()
                except (ValueError, OverflowError) as e:
                    # Set lower_slack (upper_slack below) less than -config.ecp_tolerance in this case.
                    config.logger.error(e, exc_info=True)
                    lower_slack = -10 * config.ecp_tolerance
                if lower_slack < -config.ecp_tolerance:
                    config.logger.debug(
                        'MindtPy-ECP continuing as {} has not met the '
                        'nonlinear constraints satisfaction.'
                        '\n'.format(nlc)
                    )
                    return False
            if nlc.has_ub():
                try:
                    upper_slack = nlc.uslack()
                except (ValueError, OverflowError) as e:
                    config.logger.error(e, exc_info=True)
                    upper_slack = -10 * config.ecp_tolerance
                if upper_slack < -config.ecp_tolerance:
                    config.logger.debug(
                        'MindtPy-ECP continuing as {} has not met the '
                        'nonlinear constraints satisfaction.'
                        '\n'.format(nlc)
                    )
                    return False
        # For ECP to know whether to know which bound to copy over (primal or dual)
        self.primal_bound = self.dual_bound
        config.logger.info(
            'MindtPy-ECP exiting on nonlinear constraints satisfaction. '
            'Primal Bound: {} Dual Bound: {}\n'.format(
                self.primal_bound, self.dual_bound
            )
        )

        self.best_solution_found = self.mip.clone()
        self.results.solver.termination_condition = tc.optimal
        return True
